@@ -43,8 +43,8 @@
 #define D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_FRAME_RATE_CONVERSION 0x20
 
 struct opts {
-    int deint_enabled;
-    int interlaced_only;
+    bool deint_enabled;
+    bool interlaced_only;
     int mode;
 };
 
@@ -52,7 +52,6 @@ struct priv {
     struct opts *opts;
 
     ID3D11Device *vo_dev;
-    const int *vo_formats;
 
     ID3D11DeviceContext *device_ctx;
     ID3D11VideoDevice *video_dev;
@@ -63,8 +62,6 @@ struct priv {
     D3D11_VIDEO_FRAME_FORMAT d3d_frame_format;
 
     DXGI_FORMAT out_format;
-    bool out_shared;
-    bool out_rgb;
 
     bool require_filtering;
 
@@ -99,15 +96,13 @@ static struct mp_image *alloc_pool(void *pctx, int fmt, int w, int h)
         .SampleDesc = { .Count = 1 },
         .Usage = D3D11_USAGE_DEFAULT,
         .BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE,
-        .MiscFlags = p->out_shared ? D3D11_RESOURCE_MISC_SHARED : 0,
     };
     hr = ID3D11Device_CreateTexture2D(p->vo_dev, &texdesc, NULL, &texture);
     if (FAILED(hr))
         return NULL;
 
     struct mp_image *mpi = mp_image_new_custom_ref(NULL, texture, release_tex);
-    if (!mpi)
-        abort();
+    MP_HANDLE_OOM(mpi);
 
     mp_image_setfmt(mpi, IMGFMT_D3D11);
     mp_image_set_size(mpi, w, h);
@@ -185,7 +180,7 @@ static int recreate_video_proc(struct mp_filter *vf)
         rindex = 0;
     }
 
-    // TOOD: so, how do we select which rate conversion mode the processor uses?
+    // TODO: so, how do we select which rate conversion mode the processor uses?
 
     hr = ID3D11VideoDevice_CreateVideoProcessor(p->video_dev, p->vp_enum, rindex,
                                                 &p->video_proc);
@@ -221,21 +216,9 @@ static int recreate_video_proc(struct mp_filter *vf)
     ID3D11VideoContext_VideoProcessorSetStreamColorSpace(p->video_ctx,
                                                          p->video_proc,
                                                          0, &csp);
-    if (p->out_rgb) {
-        if (p->params.color.space != MP_CSP_BT_601 &&
-            p->params.color.space != MP_CSP_BT_709)
-        {
-            MP_WARN(vf, "Unsupported video colorspace (%s/%s). Consider "
-                    "disabling hardware decoding, or using "
-                    "--hwdec=d3d11va-copy to get correct output.\n",
-                    m_opt_choice_str(mp_csp_names, p->params.color.space),
-                    m_opt_choice_str(mp_csp_levels_names, p->params.color.levels));
-        }
-    } else {
-        ID3D11VideoContext_VideoProcessorSetOutputColorSpace(p->video_ctx,
-                                                             p->video_proc,
-                                                             &csp);
-    }
+    ID3D11VideoContext_VideoProcessorSetOutputColorSpace(p->video_ctx,
+                                                         p->video_proc,
+                                                         &csp);
 
     return 0;
 fail:
@@ -350,15 +333,6 @@ cleanup:
     return out;
 }
 
-static bool vo_supports(struct priv *p, int subfmt)
-{
-    for (int n = 0; p->vo_formats && p->vo_formats[n]; n++) {
-        if (p->vo_formats[n] == subfmt)
-            return true;
-    }
-    return false;
-}
-
 static void vf_d3d11vpp_process(struct mp_filter *vf)
 {
     struct priv *p = vf->priv;
@@ -372,18 +346,8 @@ static void vf_d3d11vpp_process(struct mp_filter *vf)
         p->params = in_fmt->params;
         p->out_params = p->params;
 
-        if (vo_supports(p, IMGFMT_NV12)) {
-            p->out_params.hw_subfmt = IMGFMT_NV12;
-            p->out_format = DXGI_FORMAT_NV12;
-            p->out_shared = false;
-            p->out_rgb = false;
-        } else {
-            p->out_params.hw_subfmt = IMGFMT_RGB0;
-            p->out_format = DXGI_FORMAT_B8G8R8A8_UNORM;
-            p->out_shared = true;
-            p->out_rgb = true;
-        }
-        p->out_params.hw_flags = 0;
+        p->out_params.hw_subfmt = IMGFMT_NV12;
+        p->out_format = DXGI_FORMAT_NV12;
 
         p->require_filtering = p->params.hw_subfmt != p->out_params.hw_subfmt;
     }
@@ -463,10 +427,14 @@ static struct mp_filter *vf_d3d11vpp_create(struct mp_filter *parent,
     if (!info || !info->hwdec_devs)
         goto fail;
 
-    hwdec_devices_request_all(info->hwdec_devs);
+    struct hwdec_imgfmt_request params = {
+        .imgfmt = IMGFMT_D3D11,
+        .probing = false,
+    };
+    hwdec_devices_request_for_img_fmt(info->hwdec_devs, &params);
 
     struct mp_hwdec_ctx *hwctx =
-        hwdec_devices_get_by_lavc(info->hwdec_devs, AV_HWDEVICE_TYPE_D3D11VA);
+        hwdec_devices_get_by_imgfmt(info->hwdec_devs, IMGFMT_D3D11);
     if (!hwctx || !hwctx->av_device_ref)
         goto fail;
     AVHWDeviceContext *avhwctx = (void *)hwctx->av_device_ref->data;
@@ -475,8 +443,6 @@ static struct mp_filter *vf_d3d11vpp_create(struct mp_filter *parent,
     p->vo_dev = d3dctx->device;
     ID3D11Device_AddRef(p->vo_dev);
 
-    p->vo_formats = hwctx->supported_formats;
-
     HRESULT hr;
 
     hr = ID3D11Device_QueryInterface(p->vo_dev, &IID_ID3D11VideoDevice,
@@ -513,15 +479,15 @@ fail:
 
 #define OPT_BASE_STRUCT struct opts
 static const m_option_t vf_opts_fields[] = {
-    OPT_FLAG("deint", deint_enabled, 0),
-    OPT_FLAG("interlaced-only", interlaced_only, 0),
-    OPT_CHOICE("mode", mode, 0,
-        ({"blend", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_BLEND},
-         {"bob", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_BOB},
-         {"adaptive", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_ADAPTIVE},
-         {"mocomp", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_MOTION_COMPENSATION},
-         {"ivctc", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_INVERSE_TELECINE},
-         {"none", 0})),
+    {"deint", OPT_BOOL(deint_enabled)},
+    {"interlaced-only", OPT_BOOL(interlaced_only)},
+    {"mode", OPT_CHOICE(mode,
+        {"blend", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_BLEND},
+        {"bob", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_BOB},
+        {"adaptive", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_ADAPTIVE},
+        {"mocomp", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_MOTION_COMPENSATION},
+        {"ivctc", D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_INVERSE_TELECINE},
+        {"none", 0})},
     {0}
 };
 
@@ -531,19 +497,10 @@ const struct mp_user_filter_entry vf_d3d11vpp = {
         .name = "d3d11vpp",
         .priv_size = sizeof(OPT_BASE_STRUCT),
         .priv_defaults = &(const OPT_BASE_STRUCT) {
-            .deint_enabled = 1,
-            .interlaced_only = 0,
+            .deint_enabled = true,
             .mode = D3D11_VIDEO_PROCESSOR_PROCESSOR_CAPS_DEINTERLACE_BOB,
         },
         .options = vf_opts_fields,
     },
     .create = vf_d3d11vpp_create,
 };
-
-// Create a filter for the purpose of converting the sub-format for hwdec
-// interops which are incapable of handling some formats (ANGLE).
-struct mp_filter *vf_d3d11_create_outconv(struct mp_filter *parent)
-{
-    // options==NULL is normally not allowed, and specially handled.
-    return vf_d3d11vpp_create(parent, NULL);
-}
