@@ -19,7 +19,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <inttypes.h>
-#include <pthread.h>
 #include <limits.h>
 #include <assert.h>
 
@@ -29,16 +28,15 @@
 #include <libavutil/rational.h>
 #include <libavutil/cpu.h>
 
-#include "config.h"
-
 #include "common/msg.h"
-#include "options/m_option.h"
-#include "options/path.h"
 #include "filters/f_autoconvert.h"
 #include "filters/f_utils.h"
-#include "filters/filter.h"
 #include "filters/filter_internal.h"
+#include "filters/filter.h"
 #include "filters/user_filters.h"
+#include "options/m_option.h"
+#include "options/path.h"
+#include "osdep/threads.h"
 #include "video/img_format.h"
 #include "video/mp_image.h"
 #include "video/sws_utils.h"
@@ -54,6 +52,7 @@ struct vapoursynth_opts {
 struct priv {
     struct mp_log *log;
     struct vapoursynth_opts *opts;
+    char *script_path;
 
     VSCore *vscore;
     const VSAPI *vsapi;
@@ -62,13 +61,8 @@ struct priv {
 
     const struct script_driver *drv;
     // drv_vss
+    bool vs_initialized;
     struct VSScript *se;
-    // drv_lazy
-    struct lua_State *ls;
-    VSNodeRef **gc_noderef;
-    int num_gc_noderef;
-    VSMap **gc_map;
-    int num_gc_map;
 
     struct mp_filter *f;
     struct mp_pin *in_pin;
@@ -76,8 +70,8 @@ struct priv {
     // Format for which VS is currently configured.
     struct mp_image_params fmt_in;
 
-    pthread_mutex_t lock;
-    pthread_cond_t wakeup;
+    mp_mutex lock;
+    mp_cond wakeup;
 
     // --- the following members are all protected by lock
     struct mp_image **buffered; // oldest image first
@@ -115,25 +109,25 @@ struct script_driver {
 
 struct mpvs_fmt {
     VSPresetFormat vs;
-    int bits, cw, ch;
+    int bits, xs, ys;
 };
 
 static const struct mpvs_fmt mpvs_fmt_table[] = {
-    {pfYUV420P8,  8,  2, 2},
-    {pfYUV420P9,  9,  2, 2},
-    {pfYUV420P10, 10, 2, 2},
-    {pfYUV420P16, 16, 2, 2},
-    {pfYUV422P8,  8,  2, 1},
-    {pfYUV422P9,  9,  2, 1},
-    {pfYUV422P10, 10, 2, 1},
-    {pfYUV422P16, 16, 2, 1},
-    {pfYUV410P8,  8,  4, 4},
-    {pfYUV411P8,  8,  4, 1},
-    {pfYUV440P8,  8,  1, 2},
-    {pfYUV444P8,  8,  1, 1},
-    {pfYUV444P9,  9,  1, 1},
-    {pfYUV444P10, 10, 1, 1},
-    {pfYUV444P16, 16, 1, 1},
+    {pfYUV420P8,  8,  1, 1},
+    {pfYUV420P9,  9,  1, 1},
+    {pfYUV420P10, 10, 1, 1},
+    {pfYUV420P16, 16, 1, 1},
+    {pfYUV422P8,  8,  1, 0},
+    {pfYUV422P9,  9,  1, 0},
+    {pfYUV422P10, 10, 1, 0},
+    {pfYUV422P16, 16, 1, 0},
+    {pfYUV410P8,  8,  2, 2},
+    {pfYUV411P8,  8,  2, 0},
+    {pfYUV440P8,  8,  0, 1},
+    {pfYUV444P8,  8,  0, 0},
+    {pfYUV444P9,  9,  0, 0},
+    {pfYUV444P10, 10, 0, 0},
+    {pfYUV444P16, 16, 0, 0},
     {pfNone}
 };
 
@@ -144,7 +138,7 @@ static bool compare_fmt(int imgfmt, const struct mpvs_fmt *vs)
         return false;
     if (rfmt.component_pad > 0)
         return false;
-    if (rfmt.chroma_w != vs->cw || rfmt.chroma_h != vs->ch)
+    if (rfmt.chroma_xs != vs->xs || rfmt.chroma_ys != vs->ys)
         return false;
     if (rfmt.component_size * 8 + rfmt.component_pad != vs->bits)
         return false;
@@ -272,18 +266,12 @@ static void VS_CC vs_frame_done(void *userData, const VSFrameRef *f, int n,
 {
     struct priv *p = userData;
 
-    pthread_mutex_lock(&p->lock);
-
-    // If these assertions fail, n is an unrequested frame (or filtered twice).
-    assert(n >= p->out_frameno && n < p->out_frameno + p->max_requests);
-    int index = n - p->out_frameno;
-    MP_TRACE(p, "filtered frame %d (%d)\n", n, index);
-    assert(p->requested[index] == &dummy_img);
-
     struct mp_image *res = NULL;
     if (f) {
         struct mp_image img = map_vs_frame(p, f, false);
         struct mp_image dummy = {.params = p->fmt_in};
+        if (p->fmt_in.w != img.w || p->fmt_in.h != img.h)
+            dummy.params.crop = (struct mp_rect){0, 0, img.w, img.h};
         mp_image_copy_attributes(&img, &dummy);
         img.pkt_duration = -1;
         const VSMap *map = p->vsapi->getFramePropsRO(f);
@@ -294,11 +282,23 @@ static void VS_CC vs_frame_done(void *userData, const VSFrameRef *f, int n,
             if (!err1 && !err2)
                 img.pkt_duration = num / (double)den;
         }
-        if (img.pkt_duration < 0)
+        if (img.pkt_duration < 0) {
             MP_ERR(p, "No PTS after filter at frame %d!\n", n);
+        } else {
+            img.nominal_fps = 1.0 / img.pkt_duration;
+        }
         res = mp_image_new_copy(&img);
         p->vsapi->freeFrame(f);
     }
+
+    mp_mutex_lock(&p->lock);
+
+    // If these assertions fail, n is an unrequested frame (or filtered twice).
+    assert(n >= p->out_frameno && n < p->out_frameno + p->max_requests);
+    int index = n - p->out_frameno;
+    MP_TRACE(p, "filtered frame %d (%d)\n", n, index);
+    assert(p->requested[index] == &dummy_img);
+
     if (!res && !p->shutdown) {
         if (p->eof) {
             res = (struct mp_image *)&dummy_img_eof;
@@ -308,8 +308,8 @@ static void VS_CC vs_frame_done(void *userData, const VSFrameRef *f, int n,
         }
     }
     p->requested[index] = res;
-    pthread_cond_broadcast(&p->wakeup);
-    pthread_mutex_unlock(&p->lock);
+    mp_cond_broadcast(&p->wakeup);
+    mp_mutex_unlock(&p->lock);
     mp_filter_wakeup(p->f);
 }
 
@@ -317,7 +317,7 @@ static void vf_vapoursynth_process(struct mp_filter *f)
 {
     struct priv *p = f->priv;
 
-    pthread_mutex_lock(&p->lock);
+    mp_mutex_lock(&p->lock);
 
     if (p->failed) {
         // Not sure what we do on errors, but at least don't deadlock.
@@ -336,7 +336,7 @@ static void vf_vapoursynth_process(struct mp_filter *f)
             if (p->out_node && !p->eof) {
                 MP_VERBOSE(p, "initiate EOF\n");
                 p->eof = true;
-                pthread_cond_broadcast(&p->wakeup);
+                mp_cond_broadcast(&p->wakeup);
             }
             if (!p->out_node && mp_pin_in_needs_data(f->ppins[1])) {
                 MP_VERBOSE(p, "return EOF\n");
@@ -360,11 +360,11 @@ static void vf_vapoursynth_process(struct mp_filter *f)
                     MP_VERBOSE(p, "draining VS for format change\n");
                     mp_pin_out_unread(p->in_pin, frame);
                     p->eof = true;
-                    pthread_cond_broadcast(&p->wakeup);
+                    mp_cond_broadcast(&p->wakeup);
                     mp_filter_internal_mark_progress(f);
                     goto done;
                 }
-                pthread_mutex_unlock(&p->lock);
+                mp_mutex_unlock(&p->lock);
                 if (p->out_node)
                     destroy_vs(p);
                 p->fmt_in = mpi->params;
@@ -374,13 +374,13 @@ static void vf_vapoursynth_process(struct mp_filter *f)
                     mp_filter_internal_mark_failed(f);
                     return;
                 }
-                pthread_mutex_lock(&p->lock);
+                mp_mutex_lock(&p->lock);
             }
             if (p->out_pts == MP_NOPTS_VALUE)
                 p->out_pts = mpi->pts;
             p->frames_sent++;
             p->buffered[p->num_buffered++] = mpi;
-            pthread_cond_broadcast(&p->wakeup);
+            mp_cond_broadcast(&p->wakeup);
         } else if (frame.type != MP_FRAME_NONE) {
             MP_ERR(p, "discarding unknown frame type\n");
             mp_frame_unref(&frame);
@@ -413,7 +413,7 @@ static void vf_vapoursynth_process(struct mp_filter *f)
     if (p->requested[0] == &dummy_img_eof) {
         MP_VERBOSE(p, "finishing up\n");
         assert(p->eof);
-        pthread_mutex_unlock(&p->lock);
+        mp_mutex_unlock(&p->lock);
         destroy_vs(p);
         mp_filter_internal_mark_progress(f);
         return;
@@ -436,7 +436,7 @@ static void vf_vapoursynth_process(struct mp_filter *f)
     }
 
 done:
-    pthread_mutex_unlock(&p->lock);
+    mp_mutex_unlock(&p->lock);
 }
 
 static void VS_CC infiltInit(VSMap *in, VSMap *out, void **instanceData,
@@ -473,7 +473,7 @@ static const VSFrameRef *VS_CC infiltGetFrame(int frameno, int activationReason,
     struct priv *p = *instanceData;
     VSFrameRef *ret = NULL;
 
-    pthread_mutex_lock(&p->lock);
+    mp_mutex_lock(&p->lock);
     MP_TRACE(p, "VS asking for frame %d (at %d)\n", frameno, p->in_frameno);
     while (1) {
         if (p->shutdown) {
@@ -511,7 +511,7 @@ static const VSFrameRef *VS_CC infiltGetFrame(int frameno, int activationReason,
             // queue new frames.
             if (p->num_buffered) {
                 drain_oldest_buffered_frame(p);
-                pthread_cond_broadcast(&p->wakeup);
+                mp_cond_broadcast(&p->wakeup);
                 mp_filter_wakeup(p->f);
                 continue;
             }
@@ -535,17 +535,20 @@ static const VSFrameRef *VS_CC infiltGetFrame(int frameno, int activationReason,
                 p->vsapi->setFilterError("Could not allocate VS frame", frameCtx);
                 break;
             }
+
+            mp_mutex_unlock(&p->lock);
             struct mp_image vsframe = map_vs_frame(p, ret, true);
             mp_image_copy(&vsframe, img);
             int res = 1e6;
             int dur = img->pkt_duration * res + 0.5;
             set_vs_frame_props(p, ret, img, dur, res);
+            mp_mutex_lock(&p->lock);
             break;
         }
-        pthread_cond_wait(&p->wakeup, &p->lock);
+        mp_cond_wait(&p->wakeup, &p->lock);
     }
-    pthread_cond_broadcast(&p->wakeup);
-    pthread_mutex_unlock(&p->lock);
+    mp_cond_broadcast(&p->wakeup);
+    mp_mutex_unlock(&p->lock);
     return ret;
 }
 
@@ -553,10 +556,10 @@ static void VS_CC infiltFree(void *instanceData, VSCore *core, const VSAPI *vsap
 {
     struct priv *p = instanceData;
 
-    pthread_mutex_lock(&p->lock);
+    mp_mutex_lock(&p->lock);
     p->in_node_active = false;
-    pthread_cond_broadcast(&p->wakeup);
-    pthread_mutex_unlock(&p->lock);
+    mp_cond_broadcast(&p->wakeup);
+    mp_mutex_unlock(&p->lock);
 }
 
 // number of getAsyncFrame calls in progress
@@ -577,13 +580,13 @@ static void destroy_vs(struct priv *p)
     MP_DBG(p, "destroying VS filters\n");
 
     // Wait until our frame callbacks return.
-    pthread_mutex_lock(&p->lock);
+    mp_mutex_lock(&p->lock);
     p->initializing = false;
     p->shutdown = true;
-    pthread_cond_broadcast(&p->wakeup);
+    mp_cond_broadcast(&p->wakeup);
     while (num_requested(p))
-        pthread_cond_wait(&p->wakeup, &p->lock);
-    pthread_mutex_unlock(&p->lock);
+        mp_cond_wait(&p->wakeup, &p->lock);
+    mp_mutex_unlock(&p->lock);
 
     MP_DBG(p, "all requests terminated\n");
 
@@ -668,12 +671,20 @@ static int reinit_vs(struct priv *p, struct mp_image *input)
     struct mp_stream_info *info = mp_filter_find_stream_info(p->f);
     double container_fps = input->nominal_fps;
     double display_fps = 0;
+    int64_t display_res[2] = {0};
     if (info) {
         if (info->get_display_fps)
             display_fps = info->get_display_fps(info);
+        if (info->get_display_res) {
+            int tmp[2] = {0};
+            info->get_display_res(info, tmp);
+            display_res[0] = tmp[0];
+            display_res[1] = tmp[1];
+        }
     }
     p->vsapi->propSetFloat(vars, "container_fps", container_fps, 0);
     p->vsapi->propSetFloat(vars, "display_fps", display_fps, 0);
+    p->vsapi->propSetIntArray(vars, "display_res", display_res, 2);
 
     if (p->drv->load(p, vars) < 0)
         goto error;
@@ -688,9 +699,9 @@ static int reinit_vs(struct priv *p, struct mp_image *input)
         goto error;
     }
 
-    pthread_mutex_lock(&p->lock);
+    mp_mutex_lock(&p->lock);
     p->initializing = false;
-    pthread_mutex_unlock(&p->lock);
+    mp_mutex_unlock(&p->lock);
     MP_DBG(p, "initialized.\n");
     res = 0;
 error:
@@ -718,8 +729,8 @@ static void vf_vapoursynth_destroy(struct mp_filter *f)
     destroy_vs(p);
     p->drv->uninit(p);
 
-    pthread_cond_destroy(&p->wakeup);
-    pthread_mutex_destroy(&p->lock);
+    mp_cond_destroy(&p->wakeup);
+    mp_mutex_destroy(&p->lock);
 
     mp_filter_free_children(f);
 }
@@ -752,15 +763,14 @@ static struct mp_filter *vf_vapoursynth_create(struct mp_filter *parent,
     p->drv = p->opts->drv;
     p->f = f;
 
-    pthread_mutex_init(&p->lock, NULL);
-    pthread_cond_init(&p->wakeup, NULL);
+    mp_mutex_init(&p->lock);
+    mp_cond_init(&p->wakeup);
 
     if (!p->opts->file || !p->opts->file[0]) {
         MP_FATAL(p, "'file' parameter must be set.\n");
         goto error;
     }
-    talloc_steal(p, p->opts->file);
-    p->opts->file = mp_get_user_path(p, f->global, p->opts->file);
+    p->script_path = mp_get_user_path(p, f->global, p->opts->file);
 
     p->max_requests = p->opts->maxrequests;
     if (p->max_requests < 0)
@@ -801,15 +811,14 @@ error:
 
 #define OPT_BASE_STRUCT struct vapoursynth_opts
 static const m_option_t vf_opts_fields[] = {
-    OPT_STRING("file", file, M_OPT_FILE),
-    OPT_INTRANGE("buffered-frames", maxbuffer, 0, 1, 9999, OPTDEF_INT(4)),
-    OPT_CHOICE_OR_INT("concurrent-frames", maxrequests, 0, 1, 99,
-                      ({"auto", -1}), OPTDEF_INT(-1)),
+    {"file", OPT_STRING(file), .flags = M_OPT_FILE},
+    {"buffered-frames", OPT_INT(maxbuffer), M_RANGE(1, 9999),
+        OPTDEF_INT(4)},
+    {"concurrent-frames", OPT_CHOICE(maxrequests, {"auto", -1}),
+        M_RANGE(1, 99), OPTDEF_INT(-1)},
     {0}
 };
 
-#if HAVE_VAPOURSYNTH
-
 #include <VSScript.h>
 
 static int drv_vss_init(struct priv *p)
@@ -818,12 +827,15 @@ static int drv_vss_init(struct priv *p)
         MP_FATAL(p, "Could not initialize VapourSynth scripting.\n");
         return -1;
     }
+    p->vs_initialized = true;
     return 0;
 }
 
 static void drv_vss_uninit(struct priv *p)
 {
-    vsscript_finalize();
+    if (p->vs_initialized)
+        vsscript_finalize();
+    p->vs_initialized = false;
 }
 
 static int drv_vss_load_core(struct priv *p)
@@ -841,7 +853,7 @@ static int drv_vss_load(struct priv *p, VSMap *vars)
 {
     vsscript_setVariable(p->se, vars);
 
-    if (vsscript_evaluateFile(&p->se, p->opts->file, 0)) {
+    if (vsscript_evaluateFile(&p->se, p->script_path, 0)) {
         MP_FATAL(p, "Script evaluation failed:\n%s\n", vsscript_getError(p->se));
         return -1;
     }
@@ -868,7 +880,7 @@ static const struct script_driver drv_vss = {
 
 const struct mp_user_filter_entry vf_vapoursynth = {
     .desc = {
-        .description = "VapourSynth bridge (Python)",
+        .description = "VapourSynth bridge",
         .name = "vapoursynth",
         .priv_size = sizeof(OPT_BASE_STRUCT),
         .priv_defaults = &(const OPT_BASE_STRUCT){
@@ -878,241 +890,3 @@ const struct mp_user_filter_entry vf_vapoursynth = {
     },
     .create = vf_vapoursynth_create,
 };
-
-#endif
-
-#if HAVE_VAPOURSYNTH_LAZY
-
-#include <lua.h>
-#include <lualib.h>
-#include <lauxlib.h>
-
-#if LUA_VERSION_NUM <= 501
-#define mp_cpcall lua_cpcall
-#define FUCKYOUOHGODWHY(L) lua_pushvalue(L, LUA_GLOBALSINDEX)
-#else
-// Curse whoever had this stupid idea. Curse whoever thought it would be a good
-// idea not to include an emulated lua_cpcall() even more.
-static int mp_cpcall (lua_State *L, lua_CFunction func, void *ud)
-{
-    lua_pushcfunction(L, func); // doesn't allocate in 5.2 (but does in 5.1)
-    lua_pushlightuserdata(L, ud);
-    return lua_pcall(L, 1, 0, 0);
-}
-// Hey, let's replace old mechanisms with something slightly different!
-#define FUCKYOUOHGODWHY lua_pushglobaltable
-#endif
-
-static int drv_lazy_init(struct priv *p)
-{
-    p->ls = luaL_newstate();
-    if (!p->ls)
-        return -1;
-    luaL_openlibs(p->ls);
-    p->vsapi = getVapourSynthAPI(VAPOURSYNTH_API_VERSION);
-    p->vscore = p->vsapi ? p->vsapi->createCore(0) : NULL;
-    if (!p->vscore) {
-        MP_FATAL(p, "Could not load VapourSynth.\n");
-        lua_close(p->ls);
-        return -1;
-    }
-    return 0;
-}
-
-static void drv_lazy_uninit(struct priv *p)
-{
-    lua_close(p->ls);
-    p->vsapi->freeCore(p->vscore);
-}
-
-static int drv_lazy_load_core(struct priv *p)
-{
-    // not needed
-    return 0;
-}
-
-static struct priv *get_priv(lua_State *L)
-{
-    lua_getfield(L, LUA_REGISTRYINDEX, "p"); // p
-    struct priv *p = lua_touserdata(L, -1); // p
-    lua_pop(L, 1); // -
-    return p;
-}
-
-static void vsmap_to_table(lua_State *L, int index, VSMap *map)
-{
-    struct priv *p = get_priv(L);
-    const VSAPI *vsapi = p->vsapi;
-    for (int n = 0; n < vsapi->propNumKeys(map); n++) {
-        const char *key = vsapi->propGetKey(map, n);
-        VSPropTypes t = vsapi->propGetType(map, key);
-        switch (t) {
-        case ptInt:
-            lua_pushnumber(L, vsapi->propGetInt(map, key, 0, NULL));
-            break;
-        case ptFloat:
-            lua_pushnumber(L, vsapi->propGetFloat(map, key, 0, NULL));
-            break;
-        case ptNode: {
-            VSNodeRef *r = vsapi->propGetNode(map, key, 0, NULL);
-            MP_TARRAY_APPEND(p, p->gc_noderef, p->num_gc_noderef, r);
-            lua_pushlightuserdata(L, r);
-            break;
-        }
-        default:
-            luaL_error(L, "unknown map type");
-        }
-        lua_setfield(L, index, key);
-    }
-}
-
-static VSMap *table_to_vsmap(lua_State *L, int index)
-{
-    struct priv *p = get_priv(L);
-    const VSAPI *vsapi = p->vsapi;
-    assert(index > 0);
-    VSMap *map = vsapi->createMap();
-    MP_TARRAY_APPEND(p, p->gc_map, p->num_gc_map, map);
-    if (!map)
-        luaL_error(L, "out of memory");
-    lua_pushnil(L); // nil
-    while (lua_next(L, index) != 0) { // key value
-        if (lua_type(L, -2) != LUA_TSTRING) {
-            luaL_error(L, "key must be a string, but got %s",
-                       lua_typename(L, -2));
-        }
-        const char *key = lua_tostring(L, -2);
-        switch (lua_type(L, -1)) {
-        case LUA_TNUMBER: {
-            // gross hack because we hate everything
-            if (strncmp(key, "i_", 2) == 0) {
-                vsapi->propSetInt(map, key + 2, lua_tointeger(L, -1), 0);
-            } else {
-                vsapi->propSetFloat(map, key, lua_tonumber(L, -1), 0);
-            }
-            break;
-        }
-        case LUA_TSTRING: {
-            const char *s = lua_tostring(L, -1);
-            vsapi->propSetData(map, key, s, strlen(s), 0);
-            break;
-        }
-        case LUA_TLIGHTUSERDATA: { // assume it's VSNodeRef*
-            VSNodeRef *node = lua_touserdata(L, -1);
-            vsapi->propSetNode(map, key, node, 0);
-            break;
-        }
-        default:
-            luaL_error(L, "unknown type");
-            break;
-        }
-        lua_pop(L, 1); // key
-    }
-    return map;
-}
-
-static int l_invoke(lua_State *L)
-{
-    struct priv *p = get_priv(L);
-    const VSAPI *vsapi = p->vsapi;
-
-    VSPlugin *plugin = vsapi->getPluginByNs(luaL_checkstring(L, 1), p->vscore);
-    if (!plugin)
-        luaL_error(L, "plugin not found");
-    VSMap *map = table_to_vsmap(L, 3);
-    VSMap *r = vsapi->invoke(plugin, luaL_checkstring(L, 2), map);
-    MP_TARRAY_APPEND(p, p->gc_map, p->num_gc_map, r);
-    if (!r)
-        luaL_error(L, "?");
-    const char *err = vsapi->getError(r);
-    if (err)
-        luaL_error(L, "error calling invoke(): %s", err);
-    int err2 = 0;
-    VSNodeRef *node = vsapi->propGetNode(r, "clip", 0, &err2);
-    MP_TARRAY_APPEND(p, p->gc_noderef, p->num_gc_noderef, node);
-    if (node)
-        lua_pushlightuserdata(L, node);
-    return 1;
-}
-
-struct load_ctx {
-    struct priv *p;
-    VSMap *vars;
-    int status;
-};
-
-static int load_stuff(lua_State *L)
-{
-    struct load_ctx *ctx = lua_touserdata(L, -1);
-    lua_pop(L, 1); // -
-    struct priv *p = ctx->p;
-
-    // setup stuff; should be idempotent
-    lua_pushlightuserdata(L, p);
-    lua_setfield(L, LUA_REGISTRYINDEX, "p"); // -
-    lua_pushcfunction(L, l_invoke);
-    lua_setglobal(L, "invoke");
-
-    FUCKYOUOHGODWHY(L);
-    vsmap_to_table(L, lua_gettop(L), ctx->vars);
-    if (luaL_dofile(L, p->opts->file))
-        lua_error(L);
-    lua_pop(L, 1);
-
-    lua_getglobal(L, "video_out"); // video_out
-    if (!lua_islightuserdata(L, -1))
-        luaL_error(L, "video_out not set or has wrong type");
-    p->out_node = p->vsapi->cloneNodeRef(lua_touserdata(L, -1));
-    return 0;
-}
-
-static int drv_lazy_load(struct priv *p, VSMap *vars)
-{
-    struct load_ctx ctx = {p, vars, 0};
-    if (mp_cpcall(p->ls, load_stuff, &ctx)) {
-        MP_FATAL(p, "filter creation failed: %s\n", lua_tostring(p->ls, -1));
-        lua_pop(p->ls, 1);
-        ctx.status = -1;
-    }
-    assert(lua_gettop(p->ls) == 0);
-    return ctx.status;
-}
-
-static void drv_lazy_unload(struct priv *p)
-{
-    for (int n = 0; n < p->num_gc_noderef; n++) {
-        VSNodeRef *ref = p->gc_noderef[n];
-        if (ref)
-            p->vsapi->freeNode(ref);
-    }
-    p->num_gc_noderef = 0;
-    for (int n = 0; n < p->num_gc_map; n++) {
-        VSMap *map = p->gc_map[n];
-        if (map)
-            p->vsapi->freeMap(map);
-    }
-    p->num_gc_map = 0;
-}
-
-static const struct script_driver drv_lazy = {
-    .init = drv_lazy_init,
-    .uninit = drv_lazy_uninit,
-    .load_core = drv_lazy_load_core,
-    .load = drv_lazy_load,
-    .unload = drv_lazy_unload,
-};
-
-const struct mp_user_filter_entry vf_vapoursynth_lazy = {
-    .desc = {
-        .description = "VapourSynth bridge (Lua)",
-        .name = "vapoursynth-lazy",
-        .priv_size = sizeof(OPT_BASE_STRUCT),
-        .priv_defaults = &(const OPT_BASE_STRUCT){
-            .drv = &drv_lazy,
-        },
-        .options = vf_opts_fields,
-    },
-    .create = vf_vapoursynth_create,
-};
-
-#endif
