@@ -43,15 +43,16 @@ enum {
     VO_EVENT_AMBIENT_LIGHTING_CHANGED   = 1 << 4,
     // Special mechanism for making resizing with Cocoa react faster
     VO_EVENT_LIVE_RESIZING              = 1 << 5,
-    // Window fullscreen state changed via external influence.
-    VO_EVENT_FULLSCREEN_STATE           = 1 << 6,
+    // For VOCTRL_GET_HIDPI_SCALE changes.
+    VO_EVENT_DPI                        = 1 << 6,
     // Special thing for encode mode (vo_driver.initially_blocked).
     // Part of VO_EVENTS_USER to make vo_is_ready_for_frame() work properly.
     VO_EVENT_INITIAL_UNBLOCK            = 1 << 7,
+    VO_EVENT_FOCUS                      = 1 << 8,
 
     // Set of events the player core may be interested in.
-    VO_EVENTS_USER = VO_EVENT_RESIZE | VO_EVENT_WIN_STATE |
-                     VO_EVENT_FULLSCREEN_STATE | VO_EVENT_INITIAL_UNBLOCK,
+    VO_EVENTS_USER = VO_EVENT_RESIZE | VO_EVENT_WIN_STATE | VO_EVENT_DPI |
+                     VO_EVENT_INITIAL_UNBLOCK | VO_EVENT_FOCUS,
 };
 
 enum mp_voctrl {
@@ -65,29 +66,20 @@ enum mp_voctrl {
     VOCTRL_RESUME,
 
     VOCTRL_SET_PANSCAN,
-    VOCTRL_SET_EQUALIZER,               // struct voctrl_set_equalizer_args*
-    VOCTRL_GET_EQUALIZER,               // struct voctrl_get_equalizer_args*
+    VOCTRL_SET_EQUALIZER,
+
+    // Triggered by any change to mp_vo_opts. This is for convenience. In theory,
+    // you could install your own listener.
+    VOCTRL_VO_OPTS_CHANGED,
 
     /* private to vo_gpu */
     VOCTRL_LOAD_HWDEC_API,
 
-    // Redraw the image previously passed to draw_image() (basically, repeat
-    // the previous draw_image call). If this is handled, the OSD should also
-    // be updated and redrawn. Optional; emulated if not available.
-    VOCTRL_REDRAW_FRAME,
-
-    // Only used internally in vo_opengl_cb
+    // Only used internally in vo_libmpv
     VOCTRL_PREINIT,
     VOCTRL_UNINIT,
     VOCTRL_RECONFIG,
 
-    VOCTRL_FULLSCREEN,
-    VOCTRL_ONTOP,
-    VOCTRL_BORDER,
-    VOCTRL_ALL_WORKSPACES,
-
-    VOCTRL_GET_FULLSCREEN,
-
     VOCTRL_UPDATE_WINDOW_TITLE,         // char*
     VOCTRL_UPDATE_PLAYBACK_STATE,       // struct voctrl_playback_state*
 
@@ -95,6 +87,8 @@ enum mp_voctrl {
 
     VOCTRL_SET_CURSOR_VISIBILITY,       // bool*
 
+    VOCTRL_CONTENT_TYPE,                // enum mp_content_type*
+
     VOCTRL_KILL_SCREENSAVER,
     VOCTRL_RESTORE_SCREENSAVER,
 
@@ -103,7 +97,7 @@ enum mp_voctrl {
     VOCTRL_GET_UNFS_WINDOW_SIZE,        // int[2] (w/h)
     VOCTRL_SET_UNFS_WINDOW_SIZE,        // int[2] (w/h)
 
-    VOCTRL_GET_WIN_STATE,               // int* (VO_WIN_STATE_* flags)
+    VOCTRL_GET_FOCUSED,                 // bool*
 
     // char *** (NULL terminated array compatible with CONF_TYPE_STRING_LIST)
     // names for displays the window is on
@@ -122,28 +116,21 @@ enum mp_voctrl {
     VOCTRL_GET_ICC_PROFILE,             // bstr*
     VOCTRL_GET_AMBIENT_LUX,             // int*
     VOCTRL_GET_DISPLAY_FPS,             // double*
+    VOCTRL_GET_HIDPI_SCALE,             // double*
+    VOCTRL_GET_DISPLAY_RES,             // int[2]
+    VOCTRL_GET_WINDOW_ID,               // int64_t*
 
-    VOCTRL_GET_PREF_DEINT,              // int*
-
-    /* private to vo_gpu */
+    /* private to vo_gpu and vo_gpu_next */
     VOCTRL_EXTERNAL_RESIZE,
 };
 
-// VOCTRL_SET_EQUALIZER
-struct voctrl_set_equalizer_args {
-    const char *name;
-    int value;
+// Helper to expose what kind of content is currently playing to the VO.
+enum mp_content_type {
+    MP_CONTENT_NONE, // used for force-window
+    MP_CONTENT_IMAGE,
+    MP_CONTENT_VIDEO,
 };
 
-// VOCTRL_GET_EQUALIZER
-struct voctrl_get_equalizer_args {
-    const char *name;
-    int *valueptr;
-};
-
-// VOCTRL_GET_WIN_STATE
-#define VO_WIN_STATE_MINIMIZED 1
-
 #define VO_TRUE         true
 #define VO_FALSE        false
 #define VO_ERROR        -1
@@ -169,14 +156,12 @@ struct mp_pass_perf {
 };
 
 #define VO_PASS_PERF_MAX 64
+#define VO_PASS_DESC_MAX_LEN 128
 
 struct mp_frame_perf {
     int count;
     struct mp_pass_perf perf[VO_PASS_PERF_MAX];
-    // The owner of this struct does not have ownership over the names, and
-    // they may change at any time - so this struct should not be stored
-    // anywhere or the results reused
-    char *desc[VO_PASS_PERF_MAX];
+    char desc[VO_PASS_PERF_MAX][VO_PASS_DESC_MAX_LEN];
 };
 
 struct voctrl_performance_data {
@@ -184,7 +169,7 @@ struct voctrl_performance_data {
 };
 
 struct voctrl_screenshot {
-    bool scaled, subs, osd, high_bit_depth;
+    bool scaled, subs, osd, high_bit_depth, native_csp;
     struct mp_image *res;
 };
 
@@ -195,9 +180,17 @@ enum {
     VO_CAP_FRAMEDROP    = 1 << 1,
     // VO does not allow frames to be retained (vo_mediacodec_embed).
     VO_CAP_NORETAIN     = 1 << 2,
+    // VO supports applying film grain
+    VO_CAP_FILM_GRAIN   = 1 << 3,
+};
+
+enum {
+    // Require DR buffers to be host-cached (i.e. fast readback)
+    VO_DR_FLAG_HOST_CACHED = 1 << 0,
 };
 
 #define VO_MAX_REQ_FRAMES 10
+#define VO_MAX_SWAPCHAIN_DEPTH 8
 
 struct vo;
 struct osd_state;
@@ -213,10 +206,10 @@ struct vo_extra {
 };
 
 struct vo_frame {
-    // If > 0, realtime when frame should be shown, in mp_time_us() units.
+    // If > 0, realtime when frame should be shown, in mp_time_ns() units.
     // If 0, present immediately.
     int64_t pts;
-    // Approximate frame duration, in us.
+    // Approximate frame duration, in ns.
     int duration;
     // Realtime of estimated distance between 2 vsync events.
     double vsync_interval;
@@ -225,6 +218,10 @@ struct vo_frame {
     // "ideal" frame duration (can be different from num_vsyncs*vsync_interval
     // up to a vsync) - valid for the entire frame, i.e. not changed for repeats
     double ideal_frame_duration;
+    // "ideal" frame vsync point relative to the pts
+    double ideal_frame_vsync;
+    // "ideal" frame duration relative to the pts
+    double ideal_frame_vsync_duration;
     // how often the frame will be repeated (does not include OSD redraws)
     int num_vsyncs;
     // Set if the current frame is repeated from the previous. It's guaranteed
@@ -254,6 +251,8 @@ struct vo_frame {
     // VO if frames are dropped.
     int num_frames;
     struct mp_image *frames[VO_MAX_REQ_FRAMES];
+    // Speed unadjusted, approximate frame duration inferred from past frames
+    double approx_duration;
     // ID for frames[0] (== current). If current==NULL, the number is
     // meaningless. Otherwise, it's an unique ID for the frame. The ID for
     // a frame is guaranteed not to change (instant redraws will use the same
@@ -263,6 +262,35 @@ struct vo_frame {
     uint64_t frame_id;
 };
 
+// Presentation feedback. See get_vsync() for how backends should fill this
+// struct.
+struct vo_vsync_info {
+    // mp_time_ns() timestamp at which the last queued frame will likely be
+    // displayed (this is in the future, unless the frame is instantly output).
+    // 0 or lower if unset or unsupported.
+    // This implies the latency of the output.
+    int64_t last_queue_display_time;
+
+    // Time between 2 vsync events in nanoseconds. The difference should be the
+    // from 2 times sampled from the same reference point (it should not be the
+    // difference between e.g. the end of scanout and the start of the next one;
+    // it must be continuous).
+    // -1 if unsupported.
+    //  0 if supported, but no value available yet. It is assumed that the value
+    //    becomes available after enough swap_buffers() calls were done.
+    // >0 values are taken for granted. Very bad things will happen if it's
+    //    inaccurate.
+    int64_t vsync_duration;
+
+    // Number of skipped physical vsyncs at some point in time. Typically, this
+    // value is some time in the past by an offset that equals to the latency.
+    // This value is reset and newly sampled at every swap_buffers() call.
+    // This can be used to detect delayed frames iff you try to call
+    // swap_buffers() for every physical vsync.
+    // -1 if unset or unsupported.
+    int64_t skipped_vsyncs;
+};
+
 struct vo_driver {
     // Encoding functionality, which can be invoked via --o only.
     bool encode;
@@ -279,6 +307,9 @@ struct vo_driver {
     // Disable video timing, push frames as quickly as possible, never redraw.
     bool untimed;
 
+    // The VO is responsible for freeing frames.
+    bool frame_owner;
+
     const char *name;
     const char *description;
 
@@ -330,8 +361,10 @@ struct vo_driver {
      * allocated image. It's even possible that only 1 plane uses the buffer
      * allocated by the get_image function. The VO has to check for this.
      *
-     * stride_align is always a value >=1 that is a power of 2. The stride
-     * values of the returned image must be divisible by this value.
+     * stride_align is always a value >=1. The stride values of the returned
+     * image must be divisible by this value. This may be a non power of two.
+     *
+     * flags is a combination of VO_DR_FLAG_* flags.
      *
      * Currently, the returned image must have exactly 1 AVBufferRef set, for
      * internal implementation simplicity.
@@ -340,7 +373,7 @@ struct vo_driver {
      * will silently fallback to a default allocator
      */
     struct mp_image *(*get_image)(struct vo *vo, int imgfmt, int w, int h,
-                                  int stride_align);
+                                  int stride_align, int flags);
 
     /*
      * Thread-safe variant of get_image. Set at most one of these callbacks.
@@ -348,23 +381,13 @@ struct vo_driver {
      * vo_driver.uninit is not called before this function returns.
      */
     struct mp_image *(*get_image_ts)(struct vo *vo, int imgfmt, int w, int h,
-                                     int stride_align);
-
-    /*
-     * Render the given frame to the VO's backbuffer. This operation will be
-     * followed by a draw_osd and a flip_page[_timed] call.
-     * mpi belongs to the VO; the VO must free it eventually.
-     *
-     * This also should draw the OSD.
-     *
-     * Deprecated for draw_frame. A VO should have only either callback set.
-     */
-    void (*draw_image)(struct vo *vo, struct mp_image *mpi);
+                                     int stride_align, int flags);
 
     /* Render the given frame. Note that this is also called when repeating
      * or redrawing frames.
      *
-     * frame is freed by the caller, but the callee can still modify the
+     * frame is freed by the caller if the callee did not assume ownership
+     * of the frames, but in any case the callee can still modify the
      * contained data and references.
      */
     void (*draw_frame)(struct vo *vo, struct vo_frame *frame);
@@ -374,6 +397,15 @@ struct vo_driver {
      */
     void (*flip_page)(struct vo *vo);
 
+    /*
+     * Return presentation feedback. The implementation should not touch fields
+     * it doesn't support; the info fields are preinitialized to neutral values.
+     * Usually called once after flip_page(), but can be called any time.
+     * The values returned by this are always relative to the last flip_page()
+     * call.
+     */
+    void (*get_vsync)(struct vo *vo, struct vo_vsync_info *info);
+
     /* These optional callbacks can be provided if the GUI framework used by
      * the VO requires entering a message loop for receiving events and does
      * not call vo_wakeup() from a separate thread when there are new events.
@@ -387,7 +419,7 @@ struct vo_driver {
      * immediately.
      */
     void (*wakeup)(struct vo *vo);
-    void (*wait_events)(struct vo *vo, int64_t until_time_us);
+    void (*wait_events)(struct vo *vo, int64_t until_time_ns);
 
     /*
      * Closes driver. Should restore the original state of the system.
@@ -420,8 +452,9 @@ struct vo {
     struct mpv_global *global;
     struct vo_x11_state *x11;
     struct vo_w32_state *w32;
-    struct vo_cocoa_state *cocoa;
     struct vo_wayland_state *wl;
+    struct vo_android_state *android;
+    struct vo_drm_state *drm;
     struct mp_hwdec_devices *hwdec_devs;
     struct input_ctx *input_ctx;
     struct osd_state *osd;
@@ -479,12 +512,9 @@ void vo_query_formats(struct vo *vo, uint8_t *list);
 void vo_event(struct vo *vo, int event);
 int vo_query_and_reset_events(struct vo *vo, int events);
 struct mp_image *vo_get_current_frame(struct vo *vo);
-void vo_enable_external_renderloop(struct vo *vo);
-void vo_disable_external_renderloop(struct vo *vo);
-bool vo_render_frame_external(struct vo *vo);
-void vo_set_queue_params(struct vo *vo, int64_t offset_us, int num_req_frames);
+void vo_set_queue_params(struct vo *vo, int64_t offset_ns, int num_req_frames);
 int vo_get_num_req_frames(struct vo *vo);
-int64_t vo_get_vsync_interval(struct vo *vo);
+double vo_get_vsync_interval(struct vo *vo);
 double vo_get_estimated_vsync_interval(struct vo *vo);
 double vo_get_estimated_vsync_jitter(struct vo *vo);
 double vo_get_display_fps(struct vo *vo);
@@ -492,7 +522,7 @@ double vo_get_delay(struct vo *vo);
 void vo_discard_timing_info(struct vo *vo);
 struct vo_frame *vo_get_current_vo_frame(struct vo *vo);
 struct mp_image *vo_get_image(struct vo *vo, int imgfmt, int w, int h,
-                              int stride_align);
+                              int stride_align, int flags);
 
 void vo_wakeup(struct vo *vo);
 void vo_wait_default(struct vo *vo, int64_t until_time);
@@ -509,4 +539,6 @@ void vo_get_src_dst_rects(struct vo *vo, struct mp_rect *out_src,
 
 struct vo_frame *vo_frame_ref(struct vo_frame *frame);
 
+struct mp_image_params vo_get_current_params(struct vo *vo);
+
 #endif /* MPLAYER_VIDEO_OUT_H */
